#define _POSIX_C_SOURCE 200112L 
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <stddef.h>

#include "file_blk_dev.h"
#include "../../fs/mufs.h"
#include "../../mem/pg_alloc.h"

struct pg_zone kernel_mem_zone;

struct command {
	char *str;
	void (*fun)(struct mufs *fs, int argc, char **argv);
	int min_args;
};

void dev_create(struct file_blk_dev *dev, char *path, uint32_t size)
{
	file_blk_dev_create(dev, path, 512, size);
}

void dev_open(struct file_blk_dev *dev, char *path)
{
	struct stat buf;
	stat(path, &buf);
	file_blk_dev_create(dev, path, 512, buf.st_size);
}

void print_help()
{
	printf("usage: mufs-tool <fs-image> <command> [arg1 ... argN]\n");
	printf("commands:\n");
	printf("  mk <size in bytes> <cluster size> <inode count>\n");
	printf("  mkdir <dir>\n");
	printf("  rmdir <dir>\n");
	printf("  ls <dir>\n");
	printf("  get <file-path>\n");
	printf("  put <file-path> <local-file-path>\n");
	printf("  rm <file-path>\n");
	printf("  ln <source-path> <dest-path>\n");
	printf("  meta\n");
	printf("  meta <file-path>\n");
}

void format(char *fs_path, int argc, char **argv)
{
	if (argc < 3) {
		print_help();
		return;
	}
	uint32_t size         = atoi(argv[0]);
	uint32_t cluster_size = atoi(argv[1]);
	uint32_t inode_count  = atoi(argv[2]);
	struct file_blk_dev dev;

	dev_create(&dev, fs_path, size);
	mufs_format(&dev.blk_dev, cluster_size / 512, inode_count);
}

void ls(struct mufs *fs, int argc, char **argv)
{
	char *path = argv[0];
	struct mufs_dird *dir = mufs_dir_open(fs, argv[0], 0);
	struct dir_entry *dentry;

	while ((dentry = mufs_readdir(dir)) != NULL) {	
		char *fname = dentry->fname;
		char *full_path = malloc(strlen(path) + strlen(fname) + 2);
		strcpy(full_path, path);
		if (path[strlen(path) - 1] != '/')
			strcat(full_path, "/");
		strcat(full_path, fname);

		inode_id id   = -1;
		char type     = 'x';
		uint32_t size = -1;
		int hardlinks = -1;

		if (dentry->type == REC_DIR) {
			type = 'd';
			struct mufs_dird *d = mufs_dir_open(fs, full_path, 0);
			id = d->query.dir_inode->id;
			size = d->query.dir_inode->size;
			hardlinks = d->query.dir_inode->hardlinks; 
			mufs_dir_close(d);
		} else {
			type = 'f';
			struct mufs_fd *f = mufs_file_open(fs, full_path, 0);
			size = f->inode->size;
			id = f->inode->id;
			hardlinks = f->inode->hardlinks; 
			mufs_file_close(f);
		}
		
		printf("%c %i\t%i\t%i\t%s\n", type, id, hardlinks, size, fname);
		free(full_path);
	}
	mufs_dir_close(dir);
}

void print_bitmap_blocks(struct mufs *fs, int start_blk, int count)
{
	unsigned char buf[fs->blk_size];

	int index = 0;
	for (int i = 0; i < count; i++) {
		fs->dev->read(fs->dev, buf, start_blk + i, 1);
		for (int j = 0; j < fs->blk_size; j++) {
			for (int k = 0; k < 8; k++) {
				if (buf[j] & (1 << k)) {
					printf("%i ", index);
				}
				index++;
			}
		}
	}
}

void meta(struct mufs *fs, int argc, char **argv)
{
	if (argc == 0) {
		printf("Block size:\t%i B\n", fs->blk_size);
		printf("Supeblock size:\t%i B\n", fs->blk_size * fs->superblk_blks);
		printf("Cluster size:\t%i B\n", fs->blks_per_cluster * fs->blk_size);
		printf("Inode size:\t%i\n", (int) INODE_SIZE);
		printf("Inode blocks:\t%i blks\n", fs->inode_blks);
		printf("Inode bmap. blocks:\t%i blks\n", fs->inode_bmap_blks);
		printf("Data clusters:\t%i blks\n", fs->data_cluster_cnt);
		printf("Data bmap. blocks:\t%i blks\n", fs->data_bmap_blks);
		printf("Inodes in use: ");
		print_bitmap_blocks(fs, fs->superblk_blks + fs->inode_blks, fs->inode_bmap_blks);
		printf("\n");
		printf("Data clusters in use: ");
		print_bitmap_blocks(fs, fs->superblk_blks + fs->inode_blks + fs->inode_bmap_blks, fs->data_bmap_blks);
		printf("\n");
		return;
	}

	struct mufs_fd *f = mufs_file_open(fs, argv[0], 0);
	struct inode *inode = f->inode;
	printf("inode:\t%i\n", inode->id);
	printf("size:\t%llu B\n", inode->size);
	printf("type:\t%i\n", inode->type);
	printf("hardlnks:\t%i\n", inode->hardlinks);

	for (int i = 0; i < DIRECT_LINKS; i++) {
		printf("DIR[%i]:\t", i);
		int link = inode->direct_links[i];
		if (IS_EMPTY(link)) printf("-\n");
		else printf("%i\n", link);
	}

	if (!IS_EMPTY(inode->indirect_link)) {
		int link_cnt = (fs->blks_per_cluster * fs->blk_size) / sizeof(cluster_id);
		cluster_id links[link_cnt];
		fs->dev->read(fs->dev, (unsigned char *)links, fs->first_data_blk + (fs->blks_per_cluster * inode->indirect_link), fs->blks_per_cluster);
		for (int i = 0; i < link_cnt; i++) {
			if (IS_EMPTY(links[i])) printf (" -");
			else printf(" %i", links[i]); 
		}
		printf("\n");

	}
	mufs_file_close(f);
}


void mkdir_cmd(struct mufs *fs, int argc, char **argv)
{
	mufs_mkdir(fs, argv[0]);
}

void rmdir_cmd(struct mufs *fs, int argc, char **argv)
{
	mufs_rmdir(fs, argv[0]);
}

void put(struct mufs *fs, int argc, char **argv)
{
	unsigned char buf[128];
	FILE *f = fopen(argv[1], "rb");
	struct mufs_fd *ff = mufs_file_open(fs, argv[0], 0);
	while (!feof(f)) {
		int count = fread(buf, 1, 128, f);
		mufs_file_write(ff, buf, count);
	}
	fclose(f);
	mufs_file_close(ff);
}

void rm(struct mufs *fs, int argc, char **argv)
{
	mufs_file_unlink(fs, argv[0]);
}

void get(struct mufs *fs, int argc, char **argv)
{
	unsigned char buf[128];
	struct mufs_fd *ff = mufs_file_open(fs, argv[0], 0);
	while (!mufs_file_eof(ff)) {
		int count = mufs_file_read(ff, buf, 128);
		fwrite(buf, 1, count, stdout);
		fflush(stdout);
	}
	mufs_file_close(ff);
}

void ln(struct mufs *fs, int argc, char **argv)
{
	mufs_hardlink(fs, argv[0], argv[1]);
}

static struct command commands[] = {
	{ "ls",    ls,        1 },
	{ "meta",  meta,      0 },
	{ "put",   put,       2 },
	{ "get",   get,       1 },
	{ "ln",    ln,        2 },
	{ "rm",    rm,        1 },
	{ "mkdir", mkdir_cmd, 1 },
	{ "rmdir", rmdir_cmd, 1 },
};

#define PAGE_COUNT	(1024)
struct page pages[PAGE_COUNT];

int main(int argc, char **argv)
{
	
	if (argc < 3) {
		print_help();
		return -1;
	}

	void *mem_area;
	posix_memalign(&mem_area, PAGE_SIZE, PAGE_SIZE * PAGE_COUNT);
	pg_zone_init(&kernel_mem_zone, PAGE_COUNT, pages, mem_area);
	kmem_init();

	char *fs_path = argv[1]; 
	char *cmd     = argv[2];

	if (!strcmp(cmd, "mk")) {
		format(fs_path, argc - 3, argv + 3);
		return 0;
	}
	
	for (int i = 0; i < (sizeof(commands) / sizeof(struct command)); i++) {
		if (!strcmp(cmd, commands[i].str)) {
			if (argc - 3 < commands[i].min_args) {
				print_help();
				return -1;
			}
			struct file_blk_dev dev;
			struct mufs fs;
			dev_open(&dev, fs_path);
			mufs_open(&fs, &dev.blk_dev);
			commands[i].fun(&fs, argc - 3, argv + 3);
			return 0;
		}
	}
	print_help();
	return 0;
}
