#include <stdint.h>

#include "mufs.h"

#include "../dev/blk_dev.h"
#include "../util/string.h"
#include "../mem/alloc.h"
#include "../mem/pg_alloc.h"

/** pole, ze ktereho se berou inode pro otevrene soubory */
//static struct inode inode_buffer[MAX_OPEN_FILES];
//static struct mufs_fd fds[MAX_OPEN_FILES];

extern struct pg_zone kernel_mem_zone; 

/** hash-tabulka s aktualne otevrenymi soubory **/
static struct inode *inode_table[INODE_TABLE_SIZE];


/** pomocna funkce, ktera precte cely cluster s daty, pracuje jen v datove casti FS */
static inline int cluster_read(struct mufs *fs, uint32_t index, unsigned char *buf)
{
        return fs->dev->read(fs->dev, buf, fs->first_data_blk + (fs->blks_per_cluster * index), fs->blks_per_cluster);
}

/** pomocna funkce, ktera zapise cely cluster s daty, pracuje jen v datove casti FS */
static inline int cluster_write(struct mufs *fs, uint32_t index, unsigned char *buf)
{
        return fs->dev->write(fs->dev, buf, fs->first_data_blk + (fs->blks_per_cluster * index), fs->blks_per_cluster);
}

/** celociselne deleni se zaokrouhlovanim nahoru */
static inline uint32_t ceiling_div(uint32_t dividend, uint32_t divisor)
{
        uint32_t q = dividend / divisor;
        uint32_t m = dividend % divisor;
        if (m != 0) q ++;
        return q;
}

/**
 * pomocna funkce, ktera zarovna hodnotu val smerem nahoru 
 * na nasobek hodnoty alignment
 */
static inline uint32_t align(uint32_t val, uint32_t alignment)
{
        return ceiling_div(val, alignment) * alignment;
}

static inline int32_t div64_32(int64_t dividend, int32_t divisor)
{
        uint64_t r;
        __asm__ volatile ("idiv %2" : "=A"(r) : "A"(dividend), "r"(divisor) : "cc");
	return r & 0xffffffff;

}

static inline int32_t mod64_32(int64_t dividend, int32_t divisor)
{
        uint64_t r;
        __asm__ volatile ("idiv %2" : "=A"(r) : "A"(dividend), "r"(divisor) : "cc");
	return (r >> 32) & 0xffffffff;
}

/**
 * Najde v bitmape prvni bit, ktery je nastaveny 0. Pokud takovy bit neexistuje,
 * vraci -1;
 */
static inline int32_t find_first_zero_bit(unsigned char *bmap, int size)
{
        for (int i = 0; i < size; i++) {
                if (bmap[i] == 0xff) continue;
                for (int j = 0; j < 8; j++) {
                        if (~bmap[i] & (1 << j)) return i * 8 + j;
                }
        }
        return -1;
}

static int inode_write(struct mufs *fs, struct inode *inode);

/**
 * Vytvori strukturu souboroveho systemu na danem zarizeni.
 * blks_per_cluster -- pocet sektoru v jednom clusteru
 * inodes_cnt -- pocet inodes v souborovem systemu
 */
int mufs_format(struct blk_dev *dev, uint32_t blks_per_cluster, uint32_t inodes_cnt)
{
	unsigned char buf[SUPER_BLK_SIZE];
	struct mufs fs;

	fs.blk_size                 = dev->block_size;
	fs.superblk_blks            = SUPER_BLK_SIZE / dev->block_size; /* predpoklada, ze velikost superbloku je nasobkem velikosti sektoru */
	fs.blks_per_cluster         = blks_per_cluster;

	fs.inode_blks               = ceiling_div(inodes_cnt * INODE_SIZE, dev->block_size);
        fs.inode_bmap_blks          = ceiling_div(inodes_cnt, (dev->block_size * 8));

	uint32_t remaining_blks     = dev->block_cnt - fs.superblk_blks - fs.inode_blks - fs.inode_bmap_blks;
	uint32_t clusters_available = remaining_blks / blks_per_cluster;
	
	fs.data_bmap_blks           = ceiling_div(clusters_available, dev->block_size * 8);
        fs.data_cluster_cnt         = (remaining_blks - fs.data_bmap_blks) / blks_per_cluster;

	fs.first_data_blk           = align(fs.superblk_blks + fs.inode_blks + fs.inode_bmap_blks + fs.data_bmap_blks, blks_per_cluster);;

	/* vycisti oblast s metadaty */
	memset(buf, 0x00, SUPER_BLK_SIZE);
	for (int i = 0; i < fs.inode_blks + fs.inode_bmap_blks + fs.data_bmap_blks; i++) {
		int status = dev->write(dev, buf, fs.superblk_blks + i, 1);
		if (status) return status;
	}

	/* ulozi superblok */
	memcpy(buf, &fs, offsetof(struct mufs, dev)); /* zkopiruje perzistentni cast struktury */
	int status = dev->write(dev, buf, 0, fs.superblk_blks);
	if (status) return status;

	/* otevre souborovy system a vytvori inode s korenovym adresarem */
	mufs_open(&fs, dev);
	struct inode root_inode;
	status = mufs_dir_create(&fs, 0, &root_inode);
	root_inode.hardlinks = 2;
	inode_write(&fs, &root_inode);
	return status;
}

/**
 * Otevre existujici souborovy system na zarizeni dev
 */
int mufs_open(struct mufs *fs, struct blk_dev *dev)
{
	unsigned char superblk[SUPER_BLK_SIZE];
	if (dev->read(dev, superblk, 0, SUPER_BLK_SIZE / dev->block_size) != 0)
		return EDEV_ERR;

	/* nakopiruje perzistentni cast dat */
	memcpy(fs, superblk, sizeof(struct mufs));

	/* naplni in-memory atributy */
	fs->dev = dev;

	/* inicializuje seznam volnych inode */
/*
	fs->free_inodes = inode_buffer;
	for (int i = 0; i < MAX_OPEN_FILES - 1; i++) {
		inode_buffer[i].next = &(inode_buffer[i + 1]);
	}
	inode_buffer[MAX_OPEN_FILES - 1].next = NULL;
*/

	/* incializace struktur pro praci s otevrenymi soubory */
	fs->mem_inodes   = kmem_cache_create(sizeof(struct inode), NULL, NULL, NULL);
	fs->mem_file_des = kmem_cache_create(sizeof(struct mufs_fd), NULL, NULL, NULL);
	fs->mem_dir_des  = kmem_cache_create(sizeof(struct mufs_dird), NULL, NULL, NULL);

	for (int i = 0; i < INODE_TABLE_SIZE; i++) {
		inode_table[i] = NULL;
	}
/*
	for (int i = 0; i < MAX_OPEN_FILES; i++) {
		fds[i].inode = NULL;
	}
*/	
	return EOK;
}

/**
 * Najde prvni volny inode a oznaci jej v bitmape jako pouzity.
 */
static inode_id inode_allocate(struct mufs *fs)
{
	uint32_t blk_size = fs->blk_size;
        unsigned char bmap_blk[blk_size];

        uint32_t bmap_location = fs->superblk_blks + fs->inode_blks;
        for (uint32_t i = 0; i < fs->inode_bmap_blks; i++) {
                fs->dev->read(fs->dev, bmap_blk, bmap_location + i, 1);
                int32_t idx = find_first_zero_bit(bmap_blk, blk_size);
                if (idx >= 0) {
                        bmap_blk[idx / 8] |= 1 << (idx % 8);
                        fs->dev->write(fs->dev, bmap_blk, bmap_location + i, 1);
                        return (i * blk_size * 8) + idx;
                }
        }
        return ENO_INODES_LEFT;
}

/**
 * Oznaci inode na disku jako volny pro dalsi pouziti.
 */
static int inode_release(struct mufs *fs, inode_id inode)
{
        uint32_t blk_size = fs->blk_size;
        unsigned char bmap_blk[blk_size];
        uint32_t bmap_location = fs->superblk_blks + fs->inode_blks;

	uint32_t blk_id = (inode / (blk_size * 8));
	uint32_t idx = (inode % (blk_size * 8));

	int status = fs->dev->read(fs->dev, bmap_blk, bmap_location + blk_id, 1);
	if (status) return status;
	
	bmap_blk[idx / 8] ^= 1 << (idx % 8);
	
	return fs->dev->write(fs->dev, bmap_blk, bmap_location + blk_id, 1);
}

/**
 * Najde prvni volny cluster (TODO: vhodnejsi algoritmus)
 */
static cluster_id data_cluster_allocate(struct mufs *fs)
{
        uint32_t blk_size = fs->blk_size;
        unsigned char bmap_blk[blk_size];

        uint32_t bmap_location = fs->superblk_blks + fs->inode_blks + fs->inode_bmap_blks;
        for (uint32_t i = 0; i < fs->data_bmap_blks; i++) {
                fs->dev->read(fs->dev, bmap_blk, bmap_location + i, 1);
                int32_t idx = find_first_zero_bit(bmap_blk, blk_size);
                if (idx >= 0) {
                        bmap_blk[idx / 8] |= 1 << (idx % 8);
                        fs->dev->write(fs->dev, bmap_blk, bmap_location + i, 1);
                        return (i * blk_size * 8) + idx;
                }
        }
        return ENO_SPACE_LEFT;
}

/**
 * Oznaci datovy blok jako volny pro dalsi pouziti.
 */
static int data_cluster_release(struct mufs *fs, cluster_id cluster)
{
        uint32_t blk_size = fs->blk_size;
        unsigned char bmap_blk[blk_size];
        uint32_t bmap_location = fs->superblk_blks + fs->inode_blks + fs->inode_bmap_blks;

	uint32_t blk_id = (cluster / (blk_size * 8));
	uint32_t idx = (cluster % (blk_size * 8));

	int status = fs->dev->read(fs->dev, bmap_blk, bmap_location + blk_id, 1);
	if (status) return status;
	
	bmap_blk[idx / 8] ^= 1 << (idx % 8);
	
	return fs->dev->write(fs->dev, bmap_blk, bmap_location + blk_id, 1);
}


/**
 * Zapise dany inode na sve misto na disku.
 */
static int inode_write(struct mufs *fs, struct inode *inode)
{
        unsigned char inode_blk[fs->blk_size];

	/* index bloku na disku */
        uint32_t blk_idx = fs->superblk_blks + (inode->id / (fs->blk_size / INODE_SIZE));

	/* index na prislusnou inode v ramci daneho bloku */
        uint32_t in_blk_idx = inode->id % (fs->blk_size / INODE_SIZE);

        fs->dev->read(fs->dev, inode_blk, blk_idx, 1);

        memcpy(inode_blk + INODE_SIZE * in_blk_idx, inode, INODE_SIZE);
        return fs->dev->write(fs->dev, inode_blk, blk_idx, 1);
}

/**
 * Vytvori novy inode a zapise jej na disk.
 */
inode_id mufs_inode_create(struct mufs *fs, uint32_t type, struct inode *inode)
{
	inode_id id = inode_allocate(fs);
	if (id < 0) return id;

	/* inicializace perzistentni casti */
	memset(inode, 0x00, INODE_SIZE);
	inode->type = type;
	inode->size = 0;
	for (int i = 0; i < DIRECT_LINKS; i++) {
		inode->direct_links[i] = EMPTY_CLUSTER;
	}
	inode->indirect_link = EMPTY_CLUSTER;

	/* inicializace in-memory dat */
	inode->id = id;
	inode->active_fds = 0;

	int status = inode_write(fs, inode);
	if (status != 0) return status;

	return id;
}

/**
 * Nacte z disku hodnotu inode.
 */
int mufs_inode_read(struct mufs *fs, inode_id id, struct inode *inode)
{
        unsigned char buf[fs->blk_size];
        memset(buf, 0x00, INODE_SIZE);

	/* index bloku na disku */
        uint32_t blk_idx = fs->superblk_blks + (id / (fs->blk_size / INODE_SIZE));

	/* index zaznamu v bloku */
        uint32_t in_blk_idx = id % (fs->blk_size / INODE_SIZE);
        int status = fs->dev->read(fs->dev, buf, blk_idx, 1);
	if (status) return status;

        memcpy(inode, buf + INODE_SIZE * in_blk_idx, INODE_SIZE);

        inode->id = id;
	inode->active_fds = 0;
	return EOK;
}


/**
 * Pro dany logicky cluster urci identifikator fyzickeho clusteru
 */
static cluster_id get_physical_cluster(struct mufs *fs, struct inode *inode, cluster_id logical_cluster)
{
	if (DIRECT_LINK(logical_cluster)) return inode->direct_links[logical_cluster];
	if (INDIRECT_LINK(fs, logical_cluster)) {

		/* situace, kdy jeste nebyl vytvoreny blok s neprimymi odkazy */
		if (IS_EMPTY(inode->indirect_link))
			return EMPTY_CLUSTER;

		cluster_id links[INDIRECT_LINKS(fs)];
		if (cluster_read(fs, inode->indirect_link, (unsigned char *) links) == 0) {
			return links[logical_cluster - DIRECT_LINKS];
		}
		return EIO_ERROR;
	} else {
		return EINVALID_LINK;
	}
}

/**
 * Priradi v inode logickemu cluster cluster fyzicky.
 */
static int set_physical_cluster(struct mufs *fs, struct inode *inode, cluster_id logical_cluster, cluster_id physical_cluster)
{
	if (DIRECT_LINK(logical_cluster)) {
		int status = EOK;
		cluster_id current_cluster = inode->direct_links[logical_cluster];
		if (IS_EMPTY(physical_cluster) && !IS_EMPTY(current_cluster)) {
			status = data_cluster_release(fs, current_cluster);
		}
		inode->direct_links[logical_cluster] = physical_cluster;
		return status;
	}

	if (INDIRECT_LINK(fs, logical_cluster)) {
		cluster_id links[INDIRECT_LINKS(fs)];
		/* situace, kdy je potreba alokovat cluster neprimych odkazu */
		if (IS_EMPTY(inode->indirect_link)) {
			inode->indirect_link = data_cluster_allocate(fs);
			for (int i = 0; i < INDIRECT_LINKS(fs); i++) {
				links[i] = EMPTY_CLUSTER;
			}
		} else {
			cluster_read(fs, inode->indirect_link, (unsigned char *) links);
		}
		cluster_id current_cluster = links[logical_cluster - DIRECT_LINKS];

		if (IS_EMPTY(physical_cluster) && !IS_EMPTY(current_cluster)) {
			int status = data_cluster_release(fs, current_cluster);
			if (status) return status;
		}
		links[logical_cluster - DIRECT_LINKS] = physical_cluster;
		return cluster_write(fs, inode->indirect_link, (unsigned char *) links);
	}  else {
		return EINVALID_LINK;	
	}
}


/**
 * Precte obsah clusteru v ramci souboru, ten je dan argumentem logical_cluster.
 */
static int data_cluster_read(struct mufs *fs, struct inode *inode, cluster_id logical_cluster, unsigned char *buf)
{
	cluster_id physical_cluster = get_physical_cluster(fs, inode, logical_cluster);
	if (!IS_EMPTY(physical_cluster)) {
		return cluster_read(fs, physical_cluster, buf);
	} else {
		memset(buf, 0x00, fs->blk_size * fs->blks_per_cluster);
		return EOK;
	}
}

/**
 * Zapise cluster daneho souboru, argument logical cluster urcuje index cluster v ramci souboru 
 */
static int data_cluster_write(struct mufs *fs, struct inode *inode, cluster_id logical_cluster, unsigned char *buf)
{
	cluster_id physical_cluster = get_physical_cluster(fs, inode, logical_cluster);

	if (IS_EMPTY(physical_cluster)) {
		physical_cluster = data_cluster_allocate(fs);
		int status = set_physical_cluster(fs, inode, logical_cluster, physical_cluster);
		if (status) return status;
	}
	return cluster_write(fs, physical_cluster, buf);
}


/**
 * Zapise data do souboru specifikovaneho inode
 */
int mufs_write(struct mufs *fs, struct inode *inode, unsigned char *data, uint64_t offset, uint64_t size)
{
        uint32_t cluster_size = CLUSTER_SIZE(fs); 
	uint32_t write_size;
	while (size) {
		//cluster_id log_cluster = offset / cluster_size;
		cluster_id log_cluster = div64_32(offset, cluster_size);
		/* zapisuje/prepisuje se jen cast clusteru */
                //if ((offset % cluster_size) || (size < cluster_size)) {
                if (mod64_32(offset, cluster_size) || (size < cluster_size)) {
                        unsigned char buf[cluster_size];
                        if (data_cluster_read(fs, inode, log_cluster, buf) != 0)
				return EIO_ERROR;

			write_size = size;

			//uint32_t rel_offset = offset % cluster_size;
			uint32_t rel_offset = mod64_32(offset, cluster_size);
                        if ((size + rel_offset) > cluster_size)
				write_size = cluster_size - rel_offset;

			memcpy(buf + rel_offset, data, write_size);
                        if (data_cluster_write(fs, inode, log_cluster, buf) != 0)
				return EIO_ERROR;
		} else { /* zapisuje uplny cluster */
			write_size = cluster_size;
			if (data_cluster_write(fs, inode, log_cluster, data) != 0)
				return EIO_ERROR;
		}

		size   -= write_size;
		offset += write_size;
		data   += write_size;
	}

        if (offset > inode->size) inode->size = offset;
        return inode_write(fs, inode);
}

/**
 * Nacte do bufferu data ze souboru
 */
int mufs_read(struct mufs *fs, struct inode *inode, unsigned char *buf, uint64_t offset, uint64_t size)
{
	uint32_t cluster_size = fs->blks_per_cluster * fs->blk_size;
	unsigned char cluster[cluster_size];
	if ((offset + size) > inode->size) {
		size = inode->size - offset;
	}
	while (size > 0) {
		//if (data_cluster_read(fs, inode, offset / cluster_size, cluster) != 0)
		if (data_cluster_read(fs, inode, div64_32(offset, cluster_size), cluster) != 0)
			return EIO_ERROR;

		//uint32_t size_in_cluster = cluster_size - (offset % cluster_size);
		uint32_t size_in_cluster = cluster_size - mod64_32(offset, cluster_size);
		uint32_t read_size = size;
		if (read_size > size_in_cluster) read_size = size_in_cluster;

		memcpy(buf, cluster + mod64_32(offset, cluster_size), read_size);
		buf    += read_size;
		size   -= read_size;
		offset += read_size;
	}
	return EOK;
}

/**
 * Zkrati nebo prodlouzi soubor na danou velikost
 */
int mufs_truncate(struct mufs *fs, struct inode *inode, uint64_t size)
{
	uint32_t cluster_size = fs->blks_per_cluster * fs->blk_size;
	for (int i = ceiling_div(size, cluster_size); i < ceiling_div(inode->size, cluster_size); i++) {
		int status = set_physical_cluster(fs, inode, i, EMPTY_CLUSTER);
		if (status) return status; 
	}
	if (!IS_EMPTY(inode->indirect_link)) {
		int status = data_cluster_release(fs, inode->indirect_link);
		if (status) return status;
	}
	inode->size = size;
	return inode_write(fs, inode);
}

//
//
//
// Vysokourovnove API
//
//
//

/**
 * Pomocna funkce, ktera vraci strukturu inode.
 * Pokud je inode jiz nactena v pameti, vrati se odkaz na tato data,
 * jinak jsou data nactena z disku.
 */
static struct inode *open_inode(struct mufs *fs, inode_id id)
{
	/* zkontrolujeme, zda je soubor otevreny */
	int hash = id % INODE_TABLE_SIZE;
	struct inode *candidate = inode_table[hash];
	while (candidate) {
		if (candidate->id == id) {
			candidate->active_fds++;
			return candidate;
		}
		candidate = candidate->next;
	}

	struct inode *inode = kmem_cache_alloc(fs->mem_inodes);
	if (!inode) return NULL;

	int status = mufs_inode_read(fs, id, inode);
	if (status) return NULL;

	// TODO: sdilet kod s create_inode
//	fs->free_inodes = inode->next;
	inode->active_fds++;
	inode->next = inode_table[hash];
	inode_table[hash] = inode;	
	return inode;
}

/**
 * Ukonci praci s inode. Pokud je pocet hardlinku i aktivnich
 * deskriptoru 0, dojde k odstraneni souboru.
 */
static void close_inode(struct mufs *fs, struct inode *inode)
{
	inode->active_fds--;
	if ((inode->active_fds == 0) && (inode->hardlinks == 0)) {
		/* odstrani soubor */
		mufs_truncate(fs, inode, 0);
		inode_release(fs, inode->id);
	}

	if (inode->active_fds) return;

	int hash = inode->id % INODE_TABLE_SIZE;
	struct inode *candidate = inode_table[hash];
	if (candidate == inode) {
		inode_table[hash] = inode->next;
	} else {
		while (candidate->next != inode) {
			candidate = candidate->next;
		}
		candidate->next = inode->next;
	}

	kmem_cache_free(fs->mem_inodes, inode);
//	inode->next = fs->free_inodes;
//	fs->free_inodes = inode;
}

/**
 * Vytvori novy inode daneho typu a vrati strukturu typu inode.
 */
static struct inode *create_inode(struct mufs *fs, uint32_t type)
{
	struct inode *inode = kmem_cache_alloc(fs->mem_inodes);
	if (!inode) return NULL;

	inode_id id = mufs_inode_create(fs, type, inode);
	if (!id) return NULL;

	int hash = id % INODE_TABLE_SIZE;
	//fs->free_inodes = inode->next;
	inode->active_fds++;
	inode->next = inode_table[hash];
	inode_table[hash] = inode;	
	return inode;
}

/**
 * Vraci pro danou cestu cislo dane inode.
 * Pres parametry: file_name je mozne vratit nazev souboru,
 * parent_inode je mozne vratit cislo rodicovskeho inode.
 * Pokud tyto hodnoty nejsou potreba uvede se NULL.
 */
static inode_id path_to_inode(struct mufs *fs, char *path, char **file_name, inode_id *parent_inode)
{
	char segment[MAX_FILENAME_SIZE]; /* cast cesty */
	unsigned char buf[CLUSTER_SIZE(fs)];
	inode_id current_inode = 0;
	
	/* preskoci pripadne uvozeni zpetny lomitkem */
	while (*path && (*path == '/'))
		path++;


	char *segment_start = path;
	while (*path) {
		segment_start = path;

		int len = 0;
		char c;
		do {
			c = segment[len++] = *path;
			if (c) path++;
		} while (c && (c != '/') && (len < MAX_FILENAME_SIZE));
		if (len == MAX_FILENAME_SIZE) return EINVALID_PATH;
		if (c == '/') segment[len - 1] = '\0';


		/* otevre adresar */
		struct inode *inode = open_inode(fs, current_inode);
		if (!inode || inode->type != TYPE_DIR) {
			if (inode) close_inode(fs, inode);
			return EINVALID_PATH;
		}

		/* vyhleda soubor podle jmena */
		struct dir_query query;
		int status = mufs_dir_query(fs, &query, inode, buf, QUERY_FILE, segment);
		if (status) {
			if (file_name) *file_name = segment_start;
			close_inode(fs, inode);
			return EFILE_NOT_FOUND;
		}

		struct dir_entry *dentry = mufs_dir_next(&query);
		if (parent_inode) *parent_inode = current_inode;
		if (dentry) {
			current_inode = dentry->inode;
		} else {
			if (file_name) *file_name = segment_start;
			close_inode(fs, inode);
			return EFILE_NOT_FOUND;
		}
		close_inode(fs, inode);
	}

	if (file_name) *file_name = segment_start;
	return current_inode;
}

/**
 * Otevre soubor a vrati ukazatel do tohoto souboru.
 * Priznak flags se v tento okamzik nepouziva.
 */
struct mufs_fd *mufs_file_open(struct mufs *fs, char *path, int flags)
{
        /* vyhleda volnou strukturu */
	struct mufs_fd *fd = kmem_cache_alloc(fs->mem_file_des);
/*
	for (int i = 0; i < MAX_OPEN_FILES; i++) {
		if (!fds[i].inode) {
			fd = fds + i; 
			break;
		}
	}
*/
	if (!fd) return NULL;
	
	char *file_name;
	inode_id parent_inode; 
	struct inode *file_inode;

	inode_id inode = path_to_inode(fs, path, &file_name, &parent_inode);

	if (!strcmp(path, ".")) return NULL;
	if (!strcmp(path, "..")) return NULL;
	if ((!file_name) || (strlen(file_name) == 0) || strchr(file_name, '/')) return NULL;

	if (inode >= 0) {
		/* soubor existuje, je otevren */
		file_inode = open_inode(fs, inode);
		if (file_inode->type == TYPE_DIR) {
			close_inode(fs, file_inode);
			return NULL;
		}
	} else if (inode == EFILE_NOT_FOUND) {
		/* soubor neexistuje, je vytvoren */
		file_inode = create_inode(fs, TYPE_FILE);
		if (!file_inode) return NULL;

		struct inode *dir_inode = open_inode(fs, parent_inode);
		if (!inode) return NULL;

		int status1 = mufs_dir_link(fs, dir_inode, file_inode->id, file_name, 0);
		file_inode->hardlinks ++;

		int status2 = inode_write(fs, file_inode);
		close_inode(fs, dir_inode);
		if (status1 || status2) return NULL;
	} else {
		return NULL;
	}
	if (!file_inode) return NULL;
	file_inode->active_fds++;
	*fd = (struct mufs_fd) { fs, file_inode, 0, flags };
	return fd;
}

/**
 * Uzavre dany soubor.
 */
int mufs_file_close(struct mufs_fd *fd)
{
	close_inode(fd->fs, fd->inode);
	fd->inode = NULL;
	kmem_cache_free(fd->fs->mem_file_des, fd);
	return 0;
}

/**
 * Zapise obsah bufferu na aktualni pozici v souboru
 */
uint64_t mufs_file_write(struct mufs_fd *fd, unsigned char *buf, uint64_t size)
{
	int status = mufs_write(fd->fs, fd->inode, buf, fd->offset, size);
	fd->offset += size;
	if (status) return EIO_ERROR;
	return size;
}

/**
 * Precte do bufferu data z daneho souboru. Vraci velikosti prectenych dat.
 */
uint64_t mufs_file_read(struct mufs_fd *fd, unsigned char *buf, uint64_t size)
{
	if ((fd->offset + size) >= fd->inode->size)
		size = fd->inode->size - fd->offset;

	int status = mufs_read(fd->fs, fd->inode, buf, fd->offset, size);
	fd->offset += size;
	if (status) return EIO_ERROR;
	return size;
}

/**
 * Posune ukazatel v souboru o dany offset. Whence se chova stejne fseek v POSIX.
 */
uint64_t mufs_file_seek(struct mufs_fd *fd, int64_t offset, int whence)
{
	switch (whence) {
	case SEEK_SET: fd->offset = offset; break;
	case SEEK_CUR: fd->offset += offset; break;
	case SEEK_END: fd->offset = fd->inode->size + offset; break;
	}
	if (fd->offset < 0) fd->offset = 0;
	return fd->offset;
}

/**
 * Indikuje prekroceni hranice souboru.
 */
int mufs_file_eof(struct mufs_fd *fd)
{
	return fd->offset >= fd->inode->size;
}

/**
 * Odstrani link na dany soubor.
 */
int mufs_file_unlink(struct mufs *fs, char *path)
{
	char *file_name;
	inode_id parent;
	inode_id inode = path_to_inode(fs, path, &file_name, &parent);
	if (inode < 0) return EINVALID_PATH;
	struct inode *file_inode = open_inode(fs, inode);

	if (!file_inode) return EINVALID_PATH;
	struct inode *parent_inode = open_inode(fs, parent);
	if (!parent_inode) {
		close_inode(fs, file_inode);
		return EINVALID_PATH;
	}

	if (file_inode->type == TYPE_DIR) return EINVALID_PATH;

	mufs_dir_unlink(fs, parent_inode, file_name);
	file_inode->hardlinks --;
	inode_write(fs, file_inode);

	/* de facto necha odstranit soubor, pokud pocet odkazu klesne na 0 */
	close_inode(fs, file_inode);
	close_inode(fs, parent_inode);

	return EOK;
}

void show_free_list();
/**
 * Otevre adresar pro cteni.
 */
struct mufs_dird *mufs_dir_open(struct mufs *fs, char *path, int flags)
{
	inode_id inode = path_to_inode(fs, path, NULL, NULL);
	if (inode < 0) return NULL;

	struct inode *dir_inode = open_inode(fs, inode);

	if (!dir_inode) return NULL;

	struct mufs_dird *dird = kmem_cache_alloc(fs->mem_dir_des);
//	dird->buf = malloc(fs->blk_size * fs-> blks_per_cluster);
	dird->buf_page = pg_alloc(&kernel_mem_zone, PAGE_SIZE); // FIXME: skutecna velikost
	dird->buf = dird->buf_page->addr;
	dird->fs = fs;

	int status = mufs_dir_query(fs, &dird->query, dir_inode, dird->buf, QUERY_ALL_FILES, NULL);
	if (status) {
		//free(dird->buf);
		pg_free(&kernel_mem_zone, dird->buf_page);
		kmem_cache_free(fs->mem_dir_des, dird);
		return NULL;
	}
	return dird;
}

/**
 * Z otevreneho adresare precte jeden zaznam.
 */
struct dir_entry *mufs_readdir(struct mufs_dird *dird)
{
	return mufs_dir_next(&dird->query);
}

/**
 * Ukonci praci s otevrenym adresarem
 */
void mufs_dir_close(struct mufs_dird *dird)
{
	close_inode(dird->query.fs, dird->query.dir_inode);	
	pg_free(&kernel_mem_zone, dird->buf_page);
	kmem_cache_free(dird->fs->mem_dir_des, dird);
}

/**
 * Vytvori novy adresar
 */
int mufs_mkdir(struct mufs *fs, char *path)
{
	if (!strcmp(path, ".")) return EINVALID_PATH;
	if (!strcmp(path, "..")) return EINVALID_PATH;

	int status;
	char *dir_name;
	inode_id parent_inode;
	inode_id inode = path_to_inode(fs, path, &dir_name, &parent_inode);
	if ((!dir_name) || (strlen(dir_name) == 0) || strchr(dir_name, '/')) return EINVALID_PATH;

	if (inode >= 0) return EFILE_EXISTS;
	if (inode != EFILE_NOT_FOUND) return EINVALID_PATH;

	struct inode *dir_inode = create_inode(fs, TYPE_DIR);
	if (!dir_inode) return EIO_ERROR;

	status = mufs_dir_link(fs, dir_inode, dir_inode->id, ".", 1);
	if (status) return status;

	status = mufs_dir_link(fs, dir_inode, parent_inode, "..", 1);
	if (status) return status;	

	struct inode *parent_dir_inode = open_inode(fs, parent_inode);
	if (!parent_dir_inode) return EIO_ERROR;

	status = mufs_dir_link(fs, parent_dir_inode, dir_inode->id, dir_name, 1);
	if (status) return status;

	dir_inode->hardlinks = 2;
	parent_dir_inode->hardlinks ++;

	status = inode_write(fs, dir_inode);
	if (status) return status;

	status = inode_write(fs, parent_dir_inode);
	if (status) return status;

	close_inode(fs, parent_dir_inode);

	return EOK;
}

/**
 * Odstrani prazdny adresar
 */
int mufs_rmdir(struct mufs *fs, char *path)
{
	if (!strcmp(path, ".")) return EINVALID_PATH;
	if (!strcmp(path, "..")) return EINVALID_PATH;

	char *dir_name;
	inode_id parent_inode;
	inode_id inode = path_to_inode(fs, path, &dir_name, &parent_inode);

	struct inode *file_inode = open_inode(fs, inode);
	int hardlinks = file_inode->hardlinks;
	close_inode(fs, file_inode);

	if (hardlinks > 2) return EFILE_EXISTS;

	struct inode *dir_inode = open_inode(fs, parent_inode);
	mufs_dir_unlink(fs, dir_inode, dir_name);

	dir_inode->hardlinks--;
	int status = inode_write(fs, dir_inode);
	close_inode(fs, dir_inode);

	return status;
}


/**
 * Vytvori soubor dest_path, ktery je odkazem na soubor source_path.
 */
int mufs_hardlink(struct mufs *fs, char *source_path, char *dest_path)
{
	char *link_name;
	inode_id dest_parent_inode;

	inode_id dest_inode = path_to_inode(fs, dest_path, &link_name, &dest_parent_inode);
	if (dest_inode >= 0) return EINVALID_PATH; /* cil uz existuje */

	if ((!dest_path) || (strlen(dest_path) == 0) || strchr(dest_path, '/')) return EINVALID_PATH;

	inode_id source_inode = path_to_inode(fs, source_path, NULL, NULL);
	if (source_inode < 0) return EINVALID_PATH; /* zdroj neexistuje */

	struct inode *link_inode = open_inode(fs, source_inode);
	struct inode *dir_inode  = open_inode(fs, dest_parent_inode);

	if (!link_inode || !dir_inode) {
		if (link_inode) close_inode(fs, link_inode);
		if (dir_inode) close_inode(fs, dir_inode);
		return EINVALID_PATH;
	}

	int status = mufs_dir_link(fs, dir_inode, source_inode, link_name, link_inode->type == TYPE_DIR);
	if (status) goto close;

	link_inode->hardlinks ++;
	status = inode_write(fs, link_inode);
	

close:
	close_inode(fs, link_inode);
	close_inode(fs, dir_inode);

	return status;

}
