/*
 *
 * Implementace operaci pracujicich s adresari
 *
 */

#include <stdint.h>
#include <string.h>

#include "mufs.h"

/**
 * Vytvori inode pro adresar
 */
// TODO: presunout do format
int mufs_dir_create(struct mufs *fs, inode_id parent_dir, struct inode *dir_inode)
{
	int status;
        status = mufs_inode_create(fs, TYPE_DIR, dir_inode);
	if (status) return status;

        status = mufs_dir_link(fs, dir_inode, dir_inode->id, ".", 1);
	if (status) return status;

        status = mufs_dir_link(fs, dir_inode, parent_dir, "..", 1);
	if (status) return status;

        return EOK;
}

/**
 * Vytvori iterator, ktery umozni prochazet obsah adresare (dir_inode) s pomoci funkce:
 * mufs_dir_next.
 */
int mufs_dir_query(struct mufs *fs, struct dir_query *query, struct inode *dir_inode, unsigned char *cluster_buf, uint8_t query_type, void *constraint)
{
	uint32_t cluster_size = fs->blks_per_cluster * fs->blk_size;
	query->type = query_type;
	query->constraint.data = constraint;
	query->fs = fs;
	query->dir_inode = dir_inode;
	query->dir_offset = 0;
	query->cluster_offset = 0;
	query->cluster_buf = cluster_buf;
	if (dir_inode->size > 0) {
		return mufs_read(fs, dir_inode, cluster_buf, 0, cluster_size);
	}
	return EOK;
}

/**
 * Funkce vraci zaznam v adresari. Pozor, tento zaznam ukazuje primo
 * do bufferu, ktery byl predan do mufs_dir_query.
 */
struct dir_entry *mufs_dir_next(struct dir_query *query)
{
	uint32_t cluster_size = CLUSTER_SIZE(query->fs);
	struct dir_entry *dentry;
	while (1) {
		if (query->dir_offset >= query->dir_inode->size) {
			return NULL; /* end of dir */
		}

		/* pokud jsme za hranici clusteru, nacte dalsi cluster adresare */
		if (query->cluster_offset >= cluster_size) {
			query->dir_offset += cluster_size;
			query->cluster_offset = 0;
			mufs_read(query->fs, query->dir_inode, query->cluster_buf, query->dir_offset, cluster_size);
			continue;
		}

		dentry = (struct dir_entry *) (query->cluster_buf + query->cluster_offset);
		query->cluster_offset += dentry->rec_len;

		if ((query->type == QUERY_ALL_FILES) && (dentry->type != REC_INVALID)) return dentry;
		if ((query->type == QUERY_INVALID_RECORD) && (dentry->type == REC_INVALID) && (dentry->rec_len >= query->constraint.min_rec_size)) return dentry;
		if ((query->type == QUERY_FILE) && ((dentry->type == REC_FILE) || (dentry->type == REC_DIR)) && !strcmp(dentry->fname, query->constraint.fname)) return dentry;
	}
}

/**
 * Odstrani soubor z adresare. 
 */
int mufs_dir_unlink(struct mufs *fs, struct inode *dir_inode, char *fname) 
{
	unsigned char buf[fs->blk_size * fs->blks_per_cluster];
	struct dir_query query;
	struct dir_entry *dentry;
	
	/* testujeme, zda soubor existuje, nebo ne */
	int status = mufs_dir_query(fs, &query, dir_inode, buf, QUERY_FILE, fname);
	if (status) return status;

	dentry = mufs_dir_next(&query);
	if (dentry) {
		dentry->type = REC_INVALID;
		// TODO: record compaction
		return mufs_write(fs, dir_inode, buf, query.dir_offset, CLUSTER_SIZE(fs));
	}
	return EOK;
}

/**
 * Vytvori hardlink na soubor
 * Priznak is_dir, indikuje, jestli se jedna o hardlink na adresar.
 */
int mufs_dir_link(struct mufs *fs, struct inode *dir_inode, inode_id inode, char *fname, int is_dir)
{
	uint16_t rec_size = ALIGN4(8 + strlen(fname) + 1); /* header + jmeno + \0 */

	uint64_t dir_offset;
	unsigned char buf[CLUSTER_SIZE(fs)];

	struct dir_query query;
	struct dir_entry *dentry;
	
	/* testujeme, zda soubor existuje, nebo ne */
	int status = mufs_dir_query(fs, &query, dir_inode, buf, QUERY_FILE, fname);
	if (status) return status;

	dentry = mufs_dir_next(&query);
	if (dentry) {
		mufs_dir_unlink(fs, dir_inode, fname);
		return mufs_dir_link(fs, dir_inode, inode, fname, is_dir); 
	}

	
	/* hledame vhodne misto, kam zapsat zaznam */
	status = mufs_dir_query(fs, &query, dir_inode, buf, QUERY_INVALID_RECORD, (void *) ((unsigned long) rec_size));
	if (status) return status;

	dentry = mufs_dir_next(&query);
	if (!dentry) {
		/* pokud neni volny zaznam, alokuje cluster s jednim volnym zaznamem */
		memset(buf, 0x00, CLUSTER_SIZE(fs));
		dentry          = (struct dir_entry *) buf;
		dentry->type    = REC_INVALID;
		dentry->rec_len = CLUSTER_SIZE(fs);
		dir_offset      = dir_inode->size;
	} else {
		dir_offset = query.dir_offset;
	}	

	uint16_t space = dentry->rec_len - rec_size;

	dentry->type    = (is_dir ? REC_DIR : REC_FILE);
	dentry->inode   = inode;
	dentry->rec_len = rec_size;
	strcpy(dentry->fname, fname);

	/* doplni prazdny zaznam do volneho mista */
	if (space > 0) {
		dentry = (struct dir_entry *) (((unsigned char *) dentry) + rec_size);
		dentry->type    = REC_INVALID;
		dentry->inode   = 0;
		dentry->rec_len = space;
	}
	return mufs_write(fs, dir_inode, buf, dir_offset, CLUSTER_SIZE(fs));
}
