#include <stdint.h>
#include "dev/term.h"
#include "tasks/tasks.h"
#include "fs/mufs.h"
#include "dev/keyb.h"
#include "util/util.h"
#include "util/string.h"

#ifndef NULL
#define ((void *) 0)
#endif

#define KERNEL_END     (0x00200000)  /* hranice konce jadra */
#define USER_STACK     (0x003f0000)  /* zacatek zasobniku uzivatelskych programu */
#define MAX_ARGS       (20)          /* maximalni pocet argumentu programu */

#define STDIN          (0)
#define STDOUT         (1)

#define SYS_EXIT       (1)
#define SYS_READ       (3)
#define SYS_WRITE      (4)
#define SYS_OPEN       (5)
#define SYS_CLOSE      (6)
#define SYS_EXECVE     (11)
#define SYS_BRK        (45)
#define SYS_TRUNCATE   (92)

#define EOF	(-1)
#define EBADF	(-2)

extern struct task active_task;
extern struct mufs root_fs;

void sys_exit(int status);

/** overi, zda je adresa v rozsahu uzivatelskeho prostoru */
static inline void assert_user_address(void *addr)
{
	if ((unsigned long) addr <= KERNEL_END) {
		kprint("Segmentation fault\r\n");
		sys_exit(-1);
	}
}

void sys_exit(int status)
{
	for (int i = 0; i < MAX_FILES_OPENED; i++) {
		if (active_task.fd[i] != NULL) {
			mufs_file_close(active_task.fd[i]);
			active_task.fd[i] = NULL;
		}
	}
	active_task.last_err_code = status;
	kprint("done\r\n");
	task_exec_shell();
}


int sys_read(unsigned int fd, unsigned char *buf, unsigned int size)
{
	assert_user_address(buf);
	if ((fd < 0) || (fd > MAX_FILES_OPENED)) return EBADF;

	if (fd == STDIN) {
		int i = 0;
		while (i < size) {
			char c = keyb_getc();
			buf[i++] = c;
			term_putc(c);
			if (c == '\n') {
				term_putc('\r');
				break;
			}
		}
		return i;
	} else {
		struct mufs_fd *filedes = active_task.fd[fd];
		if (filedes == NULL) return EBADF;
		return mufs_file_read(filedes, (unsigned char *) buf, size);
	}
}

int sys_open(char *file_name, int flags, unsigned int mode)
{
	int fd = -1;
	assert_user_address(file_name);
	for (int i = STDOUT + 1; i < MAX_FILES_OPENED; i++) {
		if (active_task.fd[i] == NULL) {
			fd = i;
			break;
		}	
	}
	struct mufs_fd *filedes = mufs_file_open(&root_fs, file_name, flags);

	if (!filedes) return EBADF;

	active_task.fd[fd] = filedes;
	return fd;
}

int sys_write(unsigned int fd, unsigned char *buf, unsigned int size)
{
	assert_user_address(buf);
	if (fd == STDOUT) {
		for (unsigned int i = 0; i < size; i++) {
			if (buf[i] == '\r') continue;	
			if (buf[i] == '\n') term_putc('\r');
			term_putc(buf[i]);
		}
		return size;
	} else {
		struct mufs_fd *filedes = active_task.fd[fd];
		if (filedes == NULL) return EBADF;
		return mufs_file_write(filedes, (unsigned char *) buf, size);
	}
}

int sys_close(unsigned int fd)
{
	if (active_task.fd[fd] != NULL) {
		mufs_file_close(active_task.fd[fd]);
		active_task.fd[fd] = NULL;
		return 0;
	}
	return EBADF;
}

unsigned long sys_brk(unsigned long addr)
{
	if (addr > active_task.min_brk) {
		active_task.cur_brk = addr;
	}
	return active_task.cur_brk;
}

int sys_execve(char *pathname, char **argv, char *envp)
{
	/* ulozi argumenty pro spousteny program na zasobnik */
	unsigned long *stack        = (unsigned long *) USER_STACK;

	/* v teto casti jsou umisteny hodnoty, pred ni nasleduje oblast s odkazy */
	unsigned char *stacked_args = (unsigned char *) (stack + MAX_ARGS);
	
	int argc = 0;
	while (argv[argc] && (argc < MAX_ARGS)) {
		char *arg = argv[argc];
		int len   = strlen(arg);

		stack[argc + 1] = (unsigned long) stacked_args;

		/* prekopiruje argument na zasobnik */
		strcpy((char *) stacked_args, arg);
		stacked_args += len + 1;

		argc++;
	}

	stack[0] = argc;

	task_exec(pathname);
	return -1;
}

uint32_t syscall_handler(uint32_t syscall, uint32_t arg0, uint32_t arg1, uint32_t arg2)
{
	switch (syscall) {
	case SYS_EXIT:   sys_exit(arg0); break;
	case SYS_OPEN:   return sys_open((char *) arg0, arg1, arg2);
	case SYS_CLOSE:  return sys_close(arg0);
	case SYS_READ:   return sys_read(arg0, (unsigned char *) arg1, arg2);
	case SYS_WRITE:  return sys_write(arg0, (unsigned char *) arg1, arg2);
	case SYS_BRK:    return sys_brk(arg0);
	case SYS_EXECVE: return sys_execve((char *) arg0, (char **)arg1, (char *)arg2);
	
	default:
		kprint("unknown syscall: ");
		kprint_i(syscall);
		kprint("\r\n");
		jumpto_scheduler();
		while (1);
	}
	return 0;
}
